#  Copyright (c) 2019 Toyota Research Institute
import os

PROTOCOL_DIR = os.path.dirname(__file__)
PROCEDURE_TEMPLATE_DIR = os.path.join(PROTOCOL_DIR, "procedure_templates")
SCHEDULE_TEMPLATE_DIR = os.path.join(PROTOCOL_DIR, "schedule_templates")
PROTOCOL_SCHEMA_DIR = os.path.join(PROTOCOL_DIR, "protocol_schemas")
BIOLOGIC_TEMPLATE_DIR = os.path.join(PROTOCOL_DIR, "biologic_templates")
